@@ -379,9 +379,8 @@ FloatDivide (Value av, Value bv, int expandOk)
 
     if (FpartZero (b->mant))
     {
-	RaiseStandardException (exception_divide_by_zero,
-				"real divide by zero",
-				2, av, bv);
+	RaiseStandardException (exception_divide_by_zero, 2,
+				av, bv);
 	RETURN (Void);
     }
     DebugF ("Dividend ", a);
@@ -511,10 +510,7 @@ FloatInteger (Value av)
     }
     else
     {
-	RaiseStandardException (exception_invalid_unop_value,
-				"ambiguous conversion to int",
-				1,
-				av);
+	RaiseStandardException (exception_invalid_unop_value, 1, av);
     }
     RETURN (av);
 }
@@ -531,7 +527,11 @@ FloatFloor (Value av, int expandOk)
     if (a->exp->sign == Positive)
 	RETURN (FloatInteger (av));
     if (NaturalLess (NewNatural (a->prec), a->exp->mag))
+    {
+	if (a->mant->sign == Negative)
+	    RETURN (NewInt (-1));
 	RETURN (Zero);
+    }
     d = NaturalToInt (a->exp->mag);
     mant = FpartRsl (a->mant, d);
     if (d && a->mant->sign == Negative)
@@ -555,7 +555,11 @@ FloatCeil (Value av, int expandOk)
     if (a->exp->sign == Positive)
 	RETURN (FloatInteger (av));
     if (NaturalLess (NewNatural (a->prec), a->exp->mag))
+    {
+	if (a->mant->sign == Positive && !NaturalZero (a->mant->mag))
+	    RETURN (One);
 	RETURN (Zero);
+    }
     d = NaturalToInt (a->exp->mag);
     mant = FpartRsl (a->mant, d);
     if (d && a->mant->sign == Positive)
@@ -1154,8 +1158,9 @@ DoublePart (Value av, char *error)
     av = NewValueFloat (av, 64);
     if (!ValueIsFloat (av))
     {
-	RaiseStandardException (exception_invalid_argument, error, 
-				2, NewInt (0), av);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString (error), 
+				NewInt (0), av);
 	return 0.0;
     }
     if (NaturalLess (av->floats.exp->mag, max_int_natural))
@@ -1164,8 +1169,12 @@ DoublePart (Value av, char *error)
 	e = MAX_NICKLE_INT;
     if (e > 1023)
     {
-	RaiseStandardException (exception_invalid_argument, error,
-				2, NewInt (0), av);
+	if (av->floats.exp->sign == Negative)
+	    return 0.0;
+	
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString (error),
+				NewInt (0), av);
 	return 0.0;
     }
     if (av->floats.exp->sign == Negative)
