@@ -573,6 +573,8 @@ typedef struct _obj {
     int		used;
     int		size_stat;
     int		used_stat;
+    double_digit    ticks;
+    double_digit    sub_ticks;
     Bool	error;
     NonLocal	*nonLocal;
 } Obj;
@@ -598,10 +600,10 @@ void	    ThreadStepped (Value thread);
 void	    ThreadsWakeup (Value sleep, WakeKind wake);
 void	    ThreadsRun (Value thread, Value lex);
 void	    ThreadsInterrupt (void);
+void	    ThreadsSignal (Value signal);
 void	    ThreadsContinue (void);
 void	    ThreadFinish (Value thread, Bool error);
 void	    ThreadSetState (Value thread, ThreadState state);
-void	    ThreadClearState (Value thread, ThreadState state);
 void	    ThreadInit (void);
 void	    TraceFunction (Value file, FramePtr frame, CodePtr code, ExprPtr name);
 void	    TraceFrame (Value file, FramePtr frame, ObjPtr obj, InstPtr pc, int depth);
@@ -631,12 +633,9 @@ InstPtr	    JumpStart (Value thread, ContinuationPtr continuation, Value ret);
 JumpPtr	    NewJump (TwixtPtr leave, TwixtPtr enter, 
 		     TwixtPtr parent, ContinuationPtr continuation, Value ret);
 
-#define Runnable(t)    (((t)->thread.state & ~(ThreadSuspended)) == 0)
-
 extern Value	running;    /* current thread */
 extern Value	stopped;    /* stopped threads */
 extern Bool	complete;   /* must complete current inst */
-extern int	runnable;   /* number of non-broken threads */
 extern Bool	profiling;  /* profiling is active */
 
 void	    InstDump (InstPtr inst, int indent, int i, int *branch, int maxbranch);
@@ -700,7 +699,7 @@ void	PrettyPrint (Value f, Publish publish, SymbolPtr name);
 void	PrettyCode (Value f, CodePtr code, Atom name, Class class, 
 		    Publish publish, int level, Bool nest);
 void	PrettyStat (Value F, Expr *e, Bool nest);
-void	PrettyExpr (Value f, Expr *e, int parentPrec, int level, Bool nest, ProfileData *pd);
+void	PrettyExpr (Value f, Expr *e, int parentPrec, int level, Bool nest);
 
 void	EditFunction (SymbolPtr name, Publish publish);
 void	EditFile (Value file_name);
@@ -752,16 +751,20 @@ extern Value    yyinput;
 /* Standard exceptions */
 typedef enum _standardException {
     exception_none,
-    exception_uninitialized_value,  /* string */
+    exception_uninitialized_value,  /* */
     exception_invalid_argument,	    /* string integer poly */
-    exception_readonly_box,	    /* string poly */
-    exception_invalid_array_bounds, /* string poly poly */
-    exception_divide_by_zero,	    /* string number number */
-    exception_invalid_struct_member,/* string poly string */
-    exception_invalid_binop_values, /* string poly poly */
-    exception_invalid_unop_value,   /* string poly */
+    exception_readonly_box,	    /* poly */
+    exception_invalid_array_bounds, /* poly poly */
+    exception_divide_by_zero,	    /* number number */
+    exception_invalid_struct_member,/* poly string */
+    exception_invalid_binop_values, /* poly poly */
+    exception_invalid_unop_value,   /* poly */
     exception_open_error,	    /* string integer string */
     exception_io_error,		    /* string integer file */
+    exception_name_error,	    /* string integer string */
+    exception_signal,		    /* integer */
+    exception_system_error,	    /* string integer poly */
+    exception_io_eof,		    /* file */
     _num_standard_exceptions
 } StandardException;
 
@@ -774,7 +777,6 @@ RegisterStandardException (StandardException	se,
 
 void
 RaiseStandardException (StandardException   se,
-			char		    *string,
 			int		    argc,
 			...);
 
@@ -810,8 +812,7 @@ BoxValue (BoxPtr box, int e)
     Value   v = BoxElements(box)[e];
     if (!v)
     {
-	RaiseStandardException (exception_uninitialized_value,
-				"Uninitialized value", 0);
+	RaiseStandardException (exception_uninitialized_value, 0);
 	return (Void);
     }
     return v;
@@ -822,9 +823,8 @@ Dereference (Value v)
 {
     if (!ValueIsRef(v))
     {
-	RaiseStandardException (exception_invalid_unop_value,
-				"Not a reference",
-				1, v);
+	RaiseStandardException (exception_invalid_unop_value, 1,
+				v);
 	return Void;
     }
     return REFERENCE (RefValue (v));
@@ -912,6 +912,8 @@ Value	do_File_string_read (Value);
 Value	do_File_string_string (Value);
 Value	do_File_isatty (Value);
 Value	do_File_status (Value);
+Value	do_File_unlink (Value);
+Value	do_File_rmdir (Value);
 Value	do_String_length (Value);
 Value	do_String_new (Value);
 Value	do_Primitive_random (Value);
@@ -929,6 +931,7 @@ Value	do_profile (Value);
 
 /* two argument builtins */
 Value	do_Thread_set_priority (Value, Value);
+Value	do_Thread_signal (Value, Value);
 Value	do_File_open (Value, Value);
 Value	do_gcd (Value, Value);
 Value	do_xor (Value, Value);
@@ -940,6 +943,8 @@ Value	do_File_putc (Value, Value);
 Value	do_File_ungetb (Value, Value);
 Value	do_File_ungetc (Value, Value);
 Value	do_File_setbuf (Value, Value);
+Value	do_File_rename (Value, Value);
+Value	do_File_mkdir (Value, Value);
 Value	do_String_index (Value, Value);
 Value	do_setjmp (Value, Value);
 Value	do_setdims (Value, Value);
