@@ -12,7 +12,6 @@
 
 Value   running;
 Value   stopped;
-int	runnable;
 Bool	signalException;
 
 extern void dumpSleep (void), dumpThreads (void);
@@ -22,14 +21,17 @@ _ThreadInsert (Value thread)
 {
     Value	*prev, t;
 
-    if (Runnable (thread))
-	++runnable;
-    if (thread->thread.state == ThreadRunning)
+    switch (thread->thread.state) {
+    case ThreadRunning:
 	prev = &running;
-    else if (thread->thread.state & ThreadFinished)
-	return;
-    else
+	break;
+    case ThreadSuspended:
 	prev = &stopped;
+	break;
+    case ThreadFinished:
+    default:
+	return;
+    }
     for (; (t = *prev); prev = &t->thread.next)
 	if (t->thread.priority < thread->thread.priority)
 	    break;
@@ -42,14 +44,17 @@ _ThreadRemove (Value thread)
 {
     Value	*prev;
 
-    if (Runnable (thread))
-	--runnable;
-    if (thread->thread.state == ThreadRunning)
+    switch (thread->thread.state) {
+    case ThreadRunning:
 	prev = &running;
-    else if (thread->thread.state & ThreadFinished)
-	return;
-    else
+	break;
+    case ThreadSuspended:
 	prev = &stopped;
+	break;
+    case ThreadFinished:
+    default:
+	return;
+    }
     for (; *prev != thread; prev = &(*prev)->thread.next);
     *prev = thread->thread.next;
 }
@@ -57,17 +62,12 @@ _ThreadRemove (Value thread)
 void
 ThreadSetState (Value thread, ThreadState state)
 {
-    _ThreadRemove (thread);
-    thread->thread.state |= state;
-    _ThreadInsert (thread);
-}
-
-void
-ThreadClearState (Value thread, ThreadState state)
-{
-    _ThreadRemove (thread);
-    thread->thread.state &= ~state;
-    _ThreadInsert (thread);
+    if (state != thread->thread.state)
+    {
+	_ThreadRemove (thread);
+	thread->thread.state = state;
+	_ThreadInsert (thread);
+    }
 }
 
 void
@@ -99,11 +99,11 @@ ThreadsWakeup (Value sleep, WakeKind kind)
     for (thread = stopped; thread; thread = next)
     {
 	next = thread->thread.next;
-	if ((thread->thread.state & ThreadSuspended) && 
+	if ((thread->thread.state == ThreadSuspended) && 
 	    thread->thread.sleep == sleep)
 	{
 	    thread->thread.sleep = 0;
-	    ThreadClearState (thread, ThreadSuspended);
+	    ThreadSetState (thread, ThreadRunning);
 	    if (kind == WakeOne)
 		break;
 	}
@@ -115,53 +115,12 @@ Bool	lastThreadError;
 void
 ThreadFinish (Value thread, Bool error)
 {
-    ThreadSetState (thread, ThreadFinished);
-    ThreadsWakeup (thread, WakeAll);
-    lastThreadError = error;
-}
-
-void
-ThreadsInterrupt (void)
-{
-    Value   thread, next;
-    Value   t;
-
-    if (running)
-	t = running;
-    else
-	t = stopped;
-    for (thread = stopped; thread; thread = next)
+    if (thread->thread.state != ThreadFinished)
     {
-	next = thread->thread.next;
-	if (Runnable (thread))
-	    --runnable;
-	thread->thread.state |= ThreadInterrupted;
+	ThreadSetState (thread, ThreadFinished);
+	ThreadsWakeup (thread, WakeAll);
+	lastThreadError = error;
     }
-    for (thread = running; thread; thread = next)
-    {
-	next = thread->thread.next;
-	ThreadSetState (thread, ThreadInterrupted);
-    }
-}
-
-static int
-ThreadContinue (void)
-{
-    Value   thread, next;
-    int	    n;
-
-    n = 0;
-    for (thread = stopped; thread; thread = next)
-    {
-	next = thread->thread.next;
-
-	if (thread->thread.state & ThreadInterrupted)
-	{
-	    n++;
-	    ThreadClearState (thread, ThreadInterrupted);
-	}
-    }
-    return n;
 }
 	    
 Value
@@ -170,12 +129,12 @@ do_Thread_join (Value target)
     ENTER ();
     if (!ValueIsThread(target))
     {
-	RaiseStandardException (exception_invalid_argument,
-				"Thread::join needs thread argument",
-				2, target, Void);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString ("join needs thread argument"),
+				target, Void);
 	RETURN (Void);
     }
-    if ((target->thread.state & ThreadFinished) == 0)
+    if (target->thread.state != ThreadFinished)
     {
 	ThreadSleep (running, target, PrioritySync);
 	RETURN (Void);
@@ -186,18 +145,16 @@ do_Thread_join (Value target)
 static void
 ThreadListState (Value thread)
 {
-    int	state = thread->thread.state;
-    
-    if (state == ThreadRunning)
+    switch (thread->thread.state) {
+    case ThreadRunning:
 	FilePuts (FileStdout, " running");
-    else
-    {
-	if (state & ThreadSuspended)
-	    FilePuts (FileStdout, " suspended");
-	if (state & ThreadInterrupted)
-	    FilePuts (FileStdout, " interrupted");
-	if (state & ThreadFinished)
-	    FilePuts (FileStdout, " finished");
+	break;
+    case ThreadSuspended:
+	FilePuts (FileStdout, " suspended");
+	break;
+    case ThreadFinished:
+        FilePuts (FileStdout, " finished");
+	break;
     }
 }
 
@@ -262,9 +219,9 @@ do_Thread_set_priority (Value thread, Value priority)
     int	    i;
     if (!ValueIsThread(thread))
     {
-	RaiseStandardException (exception_invalid_argument,
-				"Thread::set_priority: not a thread",
-				2, thread, priority);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString ("set_priority: not a thread"),
+				thread, priority);
 	RETURN (Void);
     }
     i = IntPart (priority, "Invalid thread priority");
@@ -285,23 +242,14 @@ do_Thread_get_priority (Value thread)
     ENTER ();
     if (!ValueIsThread(thread))
     {
-	RaiseStandardException (exception_invalid_argument,
-				"Thread::get_priority: not a thread",
-				2, thread, Void);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString ("get_priority: not a thread"),
+				thread, Void);
 	RETURN (Void);
     }
     RETURN (NewInt (thread->thread.priority));
 }
 
-Value
-do_Thread_cont (void)
-{
-    int	    n;
-    
-    n = ThreadContinue ();
-    return NewInt (n);
-}
-
 static int
 KillThread (Value thread)
 {
@@ -309,12 +257,12 @@ KillThread (Value thread)
     
     if (!ValueIsThread(thread))
     {
-	RaiseStandardException (exception_invalid_argument,
-				"Thread::kill: not a thread",
-				2, thread, Void);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString ("kill: not a thread"),
+				thread, Void);
 	return 0;
     }
-    if (thread->thread.state & ThreadFinished)
+    if (thread->thread.state == ThreadFinished)
 	ret = 0;
     else
 	ret = 1;
@@ -333,9 +281,9 @@ do_Thread_kill (int n, Value *p)
     {
 	thread = lookupVar (0, "thread");
 	if (!ValueIsThread(thread))
-	    RaiseStandardException (exception_invalid_argument,
-				    "Thread::kill: no default thread",
-				    2, thread, Void);
+	    RaiseStandardException (exception_invalid_argument, 3,
+				    NewStrString ("kill: no default thread"),
+				    thread, Void);
 	else
 	    ret = KillThread (thread);
     }
@@ -354,7 +302,7 @@ TraceFunction (Value file, FramePtr frame, CodePtr code, ExprPtr name)
     
     FilePuts (file, "    ");
     if (name)
-	PrettyExpr (file, name, -1, 0, False, 0);
+	PrettyExpr (file, name, -1, 0, False);
     else
 	FilePuts (file, "<anonymous>");
     FilePuts (file, " (");
@@ -430,13 +378,13 @@ do_Thread_trace (int n, Value *p)
 	break;
     default:
 	if (n == 0)
-	    RaiseStandardException (exception_invalid_argument, 
-				    "Thread::trace: no default continuation",
-				    1, Zero);
+	    RaiseStandardException (exception_invalid_argument, 3,
+				    NewStrString ("trace: no default continuation"),
+				    NewInt (0), Void);
 	else
-	    RaiseStandardException (exception_invalid_argument, 
-				    "Thread::trace: neither continuation nor thread",
-				    1, v);
+	    RaiseStandardException (exception_invalid_argument, 3,
+				    NewStrString ("Thread::trace: neither continuation nor thread"),
+				    NewInt (0), v);
 	RETURN (Void);
     }
     TraceFrame (FileStdout, frame, obj, pc, depth);
@@ -894,8 +842,11 @@ JumpUnhandledException (Value thread)
 {
     Value   continuation = STACK_POP (thread->thread.continuation.stack);
     
+    /* make exec loop reschedule */
+    if (thread == running)
+	SetSignalError ();
     DebugStart (continuation);
-    SetSignalError ();
+    ThreadFinish (thread, True);
 }
 
 /*
@@ -1032,9 +983,9 @@ do_setjmp (Value continuation_ref, Value ret)
     
     if (!ValueIsRef(continuation_ref))
     {
-	RaiseStandardException (exception_invalid_argument,
-				"setjump: not a reference",
-				1, continuation_ref);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString ("setjump: not a reference"),
+				NewInt (0), continuation_ref);
 	RETURN (Void);
     }
     continuation = NewContinuation (&running->thread.continuation,
@@ -1059,9 +1010,9 @@ do_longjmp (InstPtr *next, Value continuation, Value ret)
 	RETURN (Void);
     if (!ValueIsContinuation(continuation))
     {
-	RaiseStandardException (exception_invalid_argument,
-				"longjmp: non-continuation argument",
-				1, continuation);
+	RaiseStandardException (exception_invalid_argument, 3,
+				NewStrString ("longjmp: non-continuation argument"),
+				NewInt (0), continuation);
 	RETURN (Void);
     }
     RETURN (ContinuationJump (running, &continuation->continuation, ret, next));
@@ -1097,7 +1048,6 @@ TwixtMark (void *object)
     TwixtPtr	twixt = object;
 
     ContinuationMark (&twixt->continuation);
-    MemReference (twixt->leave);
 }
 
 DataType    TwixtType = { TwixtMark, 0, "TwixtType" };
@@ -1221,7 +1171,6 @@ CheckStandardException (void)
 
 void
 RaiseStandardException (StandardException   se,
-			char		    *string,
 			int		    argc,
 			...)
 {
@@ -1231,12 +1180,10 @@ RaiseStandardException (StandardException   se,
     va_list	va;
     
     va_start (va, argc);
-    i = argc + 1;
+    i = argc;
     args = NewArray (False, False, typePoly, 1, &i);
-    ArrayValueSet (&args->array, 0, NewStrString (string));
-    if (argc)
-	for (i = 0; i < argc; i++)
-	    ArrayValueSet (&args->array, i + 1, va_arg (va, Value));
+    for (i = 0; i < argc; i++)
+	ArrayValueSet (&args->array, i, va_arg (va, Value));
     standardException = se;
     standardExceptionArgs = args;
     SetSignalException ();
@@ -1257,3 +1204,61 @@ JumpStandardException (Value thread, InstPtr *next)
     standardExceptionArgs = 0;
     RETURN (args);
 }
+
+static void
+SignalThread (Value thread, Value signal, Bool executing)
+{
+    ENTER ();
+    int		i = 1;
+    Value	args = NewArray (False, False, typePoly, 1, &i);
+    SymbolPtr	except = standardExceptions[exception_signal];
+    
+    ArrayValueSet (&args->array, 0, signal);
+    if (thread == running && executing)
+    {
+	standardException = exception_signal;
+	standardExceptionArgs = args;
+	SetSignalException ();
+    }
+    else if (except)
+    {
+	InstPtr	next;
+	
+	RaiseException (thread, except, args, &next);
+	thread->thread.continuation.value = args;
+	thread->thread.continuation.pc = next;
+	if (thread->thread.state == ThreadSuspended) {
+	    thread->thread.sleep = 0;
+	    ThreadSetState (thread, ThreadRunning);
+	}
+    }
+    EXIT ();
+}
+
+void
+ThreadsSignal (Value signal)
+{
+    ENTER ();
+    Value   thread, next;
+
+    /* do running first -- signalling makes threads run */
+    for (thread = running; thread; thread = next)
+    {
+	next = thread->thread.next;
+	SignalThread (thread, signal, False);
+    }
+    for (thread = stopped; thread; thread = next)
+    {
+	next = thread->thread.next;
+	SignalThread (thread, signal,  False);
+    }
+    EXIT ();
+}
+
+Value
+do_Thread_signal (Value thread, Value signal)
+{
+    ENTER ();
+    SignalThread (thread, signal, True);
+    RETURN (Void);
+}
